use std::fmt::{Display, Formatter, Result};

use eframe::App;
use egui::{
    vec2, CentralPanel, ComboBox, Context, Frame, Pos2, Rect, Sense, Shape, Slider, TextStyle,
    TopBottomPanel,
};
use strum::IntoEnumIterator;
use strum_macros::EnumIter;

use crate::game::Universe;

const CANVAS_WIDTH: f32 = 1280.0;
const CANVAS_HEIGHT: f32 = 720.0;

#[repr(u8)]
#[derive(Clone, Copy, Debug, PartialEq, Eq, EnumIter)]
enum Speed {
    Slow = 8,
    Normal = 4,
    Fast = 2,
    Fastest = 1,
}

impl Default for Speed {
    fn default() -> Self {
        Speed::Fast
    }
}

impl Display for Speed {
    fn fmt(&self, f: &mut Formatter) -> Result {
        match self {
            Speed::Slow => write!(f, "Slow"),
            Speed::Normal => write!(f, "Normal"),
            Speed::Fast => write!(f, "Fast"),
            Speed::Fastest => write!(f, "Fastest"),
        }
    }
}

#[repr(u32)]
#[derive(Clone, Copy, Debug, PartialEq, Eq, EnumIter)]
enum Size {
    Small = 8,
    Medium = 4,
    Large = 2,
    Giant = 1,
}

impl Default for Size {
    fn default() -> Self {
        Size::Large
    }
}

impl Display for Size {
    fn fmt(&self, f: &mut Formatter) -> Result {
        match self {
            Size::Small => write!(f, "Small"),
            Size::Medium => write!(f, "Medium"),
            Size::Large => write!(f, "Large"),
            Size::Giant => write!(f, "Giant"),
        }
    }
}

pub struct MyEguiApp {
    game: Universe,
    frame_counter: u8,
    speed: Speed,
    size: Size,
    cache: Vec<Shape>,
    paused: bool,
    click: Option<Pos2>,
    life_level: i32,
}

impl MyEguiApp {
    pub fn new(_cc: &eframe::CreationContext<'_>) -> Self {
        let size = Size::default();
        Self {
            game: Universe::new(
                CANVAS_WIDTH as u32 / size as u32,
                CANVAS_HEIGHT as u32 / size as u32,
                2,
            ),
            frame_counter: 0,
            speed: Speed::default(),
            size,
            cache: Vec::new(),
            paused: false,
            click: None,
            life_level: 2,
        }
    }
}

impl App for MyEguiApp {
    fn update(&mut self, ctx: &Context, frame: &mut eframe::Frame) {
        let toolbar = self.render_toolbar(ctx);
        #[cfg(not(target_arch = "wasm32"))]
        frame.set_window_size(vec2(CANVAS_WIDTH, CANVAS_HEIGHT + toolbar.height()));
        self.render_canvas(ctx);
        if !self.paused {
            self.frame_counter += 1;
            self.frame_counter %= self.speed as u8;
        }
        ctx.request_repaint();
    }
}

impl MyEguiApp {
    fn reset(&mut self) {
        self.game = Universe::new(
            CANVAS_WIDTH as u32 / self.size as u32,
            CANVAS_HEIGHT as u32 / self.size as u32,
            self.life_level,
        );
        self.cache = Vec::new();
        self.frame_counter = 0;
    }

    fn render_toolbar(&mut self, ctx: &Context) -> Rect {
        let mut frame = egui::Frame::side_top_panel(ctx.style().as_ref());
        frame.inner_margin.bottom += 1.5;
        let toolbar = TopBottomPanel::top("my_panel")
            .frame(frame)
            .show(ctx, |ui| {
                ui.horizontal(|ui| {
                    ui.style_mut().spacing.interact_size.y *= 1.25;
                    ui.style_mut()
                        .text_styles
                        .get_mut(&TextStyle::Button)
                        .unwrap()
                        .size *= 1.25;
                    ui.add(Slider::new(&mut self.life_level, 0..=20).text("Life level"));
                    if ui.button("Reset").clicked() {
                        self.reset();
                    }
                    let paused = self.paused;
                    ui.toggle_value(&mut self.paused, if paused { "▶" } else { "⏸" });
                    ComboBox::from_label("Speed")
                        .selected_text(format!("{:?}", self.speed))
                        .show_ui(ui, |ui| {
                            for speed in Speed::iter() {
                                if ui
                                    .selectable_value(&mut self.speed, speed, speed.to_string())
                                    .clicked()
                                {
                                    self.frame_counter = 0;
                                }
                            }
                        });
                    ComboBox::from_label("Size")
                        .selected_text(format!("{:?}", self.size))
                        .show_ui(ui, |ui| {
                            for size in Size::iter() {
                                if ui
                                    .selectable_value(&mut self.size, size, size.to_string())
                                    .clicked()
                                {
                                    self.reset();
                                }
                            }
                        });
                });
            });
        toolbar.response.rect
    }

    fn render_canvas(&mut self, ctx: &Context) -> Rect {
        let frame = Frame::central_panel(ctx.style().as_ref()).inner_margin(0.0);
        let canvas = CentralPanel::default().frame(frame).show(ctx, |ui| {
            let (res, painter) =
                ui.allocate_painter(vec2(CANVAS_WIDTH, CANVAS_HEIGHT), Sense::click());
            let canvas_origin = res.rect.left_top().to_vec2();
            let interaction = match (res.interact_pointer_pos(), self.click) {
                (Some(pointer_pos), None) => {
                    let canvas_pos = pointer_pos - canvas_origin;
                    self.click = Some(canvas_pos);
                    let x = canvas_pos.x as u32 / self.size as u32;
                    let y = canvas_pos.y as u32 / self.size as u32;
                    self.game.toggle_cell(y, x);
                    true
                }
                (None, Some(_)) => {
                    self.click = None;
                    false
                }
                _ => false,
            };
            let tick = !self.paused && self.frame_counter % self.speed as u8 == 0;
            if tick {
                self.game.tick();
            }
            if tick || interaction || self.cache.is_empty() {
                self.cache = self
                    .game
                    .render(canvas_origin, vec2(CANVAS_WIDTH, CANVAS_HEIGHT));
            }
            painter.extend(self.cache.iter().cloned());
        });
        canvas.response.rect
    }
}
