use eframe::epaint::RectShape;
use egui::{pos2, vec2, Color32, Rect, Rounding, Shape, Stroke, Vec2};
use rand::Rng;

#[repr(u8)]
#[derive(Clone, Copy, Debug, PartialEq, Eq)]
pub enum Cell {
    Dead = 0,
    Alive = 1,
}

pub struct Universe {
    width: u32,
    height: u32,
    age: u32,
    cells: (Vec<Cell>, Vec<Cell>),
}

impl Universe {
    pub fn new(width: u32, height: u32, life_level: i32) -> Universe {
        let mut rng = rand::thread_rng();
        let cells = (0..width * height)
            .map(|_| {
                if rng.gen_range(0..=20) < life_level {
                    Cell::Alive
                } else {
                    Cell::Dead
                }
            })
            .collect::<Vec<Cell>>();

        Universe {
            width,
            height,
            age: 0,
            cells: (cells.clone(), cells),
        }
    }

    fn render_rect(
        canvas_origin: Vec2,
        y: f32,
        x: f32,
        width_counter: u32,
        cell_side: f32,
    ) -> Shape {
        const FILL: Color32 = Color32::from_gray(128);
        const ROUNDING: Rounding = Rounding {
            nw: 0.0,
            ne: 0.0,
            sw: 0.0,
            se: 0.0,
        };
        const STROKE: Stroke = Stroke::NONE;
        let rect = Rect::from_min_size(
            pos2(x, y) + canvas_origin,
            vec2(cell_side * width_counter as f32, cell_side),
        );
        Shape::Rect(RectShape {
            rect,
            rounding: ROUNDING,
            fill: FILL,
            stroke: STROKE,
        })
    }

    pub fn render(&self, canvas_origin: Vec2, canvas_size: Vec2) -> Vec<Shape> {
        let cell_side = canvas_size.x / self.width as f32;
        let mut rectangles_cache = Vec::new();

        let current = match self.age % 2 {
            0 => &self.cells.1,
            _ => &self.cells.0,
        };
        let mut width_counter = 0;
        let mut width_origin = 0.0;
        for row in 0..self.height {
            for col in 0..self.width {
                let idx = Universe::get_index(self.width, row, col);
                if let Cell::Alive = current[idx] {
                    if width_counter == 0 {
                        width_origin = col as f32 * cell_side;
                    }
                    width_counter += 1;
                } else if width_counter != 0 {
                    rectangles_cache.push(Universe::render_rect(
                        canvas_origin,
                        row as f32 * cell_side,
                        width_origin,
                        width_counter,
                        cell_side,
                    ));
                    width_counter = 0;
                }
            }
            if width_counter != 0 {
                rectangles_cache.push(Universe::render_rect(
                    canvas_origin,
                    row as f32 * cell_side,
                    width_origin,
                    width_counter,
                    cell_side,
                ));
                width_counter = 0;
            }
        }
        rectangles_cache
    }

    fn get_index(width: u32, row: u32, column: u32) -> usize {
        (row * width + column) as usize
    }

    fn live_neighbor_count(
        universe: &[Cell],
        height: u32,
        width: u32,
        row: u32,
        column: u32,
    ) -> u8 {
        let mut count = 0;
        for delta_row in [height - 1, 0, 1] {
            for delta_col in [width - 1, 0, 1] {
                if delta_row == 0 && delta_col == 0 {
                    continue;
                }

                let neighbor_row = (row + delta_row) % height;
                let neighbor_col = (column + delta_col) % width;
                let idx = Universe::get_index(width, neighbor_row, neighbor_col);
                count += universe[idx] as u8;
            }
        }
        count
    }

    pub fn tick(&mut self) {
        let (next, current) = match self.age % 2 {
            0 => (&mut self.cells.0, &self.cells.1),
            _ => (&mut self.cells.1, &self.cells.0),
        };
        for row in 0..self.height {
            for col in 0..self.width {
                let idx = Universe::get_index(self.width, row, col);
                let cell = current[idx];
                let live_neighbors =
                    Universe::live_neighbor_count(current, self.height, self.width, row, col);

                let next_cell = match (cell, live_neighbors) {
                    // Rule 1: Any live cell with fewer than two live neighbours
                    // dies, as if caused by underpopulation.
                    (Cell::Alive, x) if x < 2 => Cell::Dead,
                    // Rule 2: Any live cell with two or three live neighbours
                    // lives on to the next generation.
                    (Cell::Alive, 2) | (Cell::Alive, 3) => Cell::Alive,
                    // Rule 3: Any live cell with more than three live
                    // neighbours dies, as if by overpopulation.
                    (Cell::Alive, x) if x > 3 => Cell::Dead,
                    // Rule 4: Any dead cell with exactly three live neighbours
                    // becomes a live cell, as if by reproduction.
                    (Cell::Dead, 3) => Cell::Alive,
                    // All other cells remain in the same state.
                    (otherwise, _) => otherwise,
                };

                next[idx] = next_cell;
            }
        }
        self.age += 1;
    }

    pub fn toggle_cell(&mut self, row: u32, column: u32) {
        let idx = Universe::get_index(self.width, row, column);
        let current = match self.age % 2 {
            0 => &mut self.cells.1,
            _ => &mut self.cells.0,
        };
        current[idx] = match current[idx] {
            Cell::Dead => Cell::Alive,
            Cell::Alive => Cell::Dead,
        };
    }
}
